/////////////////////////////////////////////////////////////////////////////////////////////
///
///
///	\file	TBaseClient.h
///	\author	Rob Bateman
/// \brief	this file defines the lowest level of Pipe IO as the template TBaseClient. The
///			template argument should be a function object providing a single () operator
///			that accepts a std::istream object. This will be called whenever the pipe
///			reads some data. 
///			
///	
///

/////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////
//
//	Includes
//

#include <iostream>
#include <strstream>
#include <string>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>



/////////////////////////////////////////////////////////////////////////////////////////////
//
//	Constants
//

const unsigned int STRING_SIZE = 256;
const unsigned int BUFFER_SIZE = 1000;



/////////////////////////////////////////////////////////////////////////////////////////////
//
//	Base Client template
//

template
< 
	typename OnRecieveData
>
class TBaseClient
{
private:

	///	File or Pipe handle.
	static HANDLE 	  ms_hPipe;
	///	Overlapped structure
	static OVERLAPPED ms_OverLapWrt;
	///	Event handle for overlapped writes.
	static HANDLE     ms_hEventWrt;
	
	/// This holds the name of the server
	static char ms_pServerName[ STRING_SIZE ];
	///	This holds the name of the client pipe
	static char ms_pClientName[ STRING_SIZE ];
	
	///	an internal buffer (lets use heap allocated memory)
	static char ms_pOutBuffer [ BUFFER_SIZE ];
	///	an internal buffer (lets use heap allocated memory)
	static char ms_pInBuffer  [ BUFFER_SIZE ];
	
	///	The function object for recieving data
	static OnRecieveData OnRead;
	
	///	\brief	this structure is a wrapper for the std::strstream objects
	///	
	struct _In {
		///	\param	dat	-	the data buffer
		///	\param	sz	-	size of buffer
		///
		_In(char *dat,unsigned int sz) 
			: m_Buffer(dat,sz) {}
		
		///	the input string buffer
		std::istrstream m_Buffer; 
	};
	
	///	\brief	this structure is a wrapper for the std::strstream objects
	///	
	struct _Out {
		///	\param	dat	-	the data buffer
		///	\param	sz	-	size of buffer
		///
		_Out(char *dat,unsigned int sz) 
			: m_Buffer(dat,sz) {}
		
		///	the output string buffer
		std::ostrstream m_Buffer; 
	};
	
	///	a utility stream....
	static _In  *ms_pIn; 
	static _Out *ms_pOut; 
	
	
	///	\brief	this function resets the input stream
	///
	static inline void ResetInputBuffer() {
		delete ms_pIn;
		ms_pIn = new _In( ms_pInBuffer, BUFFER_SIZE );
	}
	
	///	\brief	this is the thread procudure
	///	\param	a handle to the thread
	///
	static VOID __stdcall ReadPipe(HANDLE *hPipe);

	
protected:

	///	\brief	this function returns a reference to the internal ostrstream object.
	///			this can be used to specify send data usin normal C++ streams (cin,cout etc)
	///	\return	a reference to the stream
	///
	static inline std::ostrstream &BufferStream() {
		if(!ms_pOut)
			ms_pOut = new _Out( ms_pOutBuffer, BUFFER_SIZE ); 
		return ms_pOut->m_Buffer;
	}
	
	///	\brief	this function will send the data currently stored in the buffer
	///	\return	true if OK
	///
	static bool Send();

public:

	/// \brief	this function retrieves the name of the server pipe
	///	\return	the name of the server pipe
	///
	static inline std::string GetServerName() {
		return std::string( ms_pServerName );
	}

	///	\brief	this function returns the name of the client
	///	\return	the name of the client
	///
	static inline std::string GetClientName() {
		return std::string( ms_pClientName );
	}
	
	///	\brief	this function connects to the specified server
	///	\param	ServerName	-	the name of the server to connect to (. for local)
	///	\return	true if PIPE found
	///
	static bool Connect( std::string ClientName, std::string ServerName = "." );

	///	\brief	this function breaks the connection to the server
	///	\return	true if broken
	///
	static bool Disconnect();
	
};



/////////////////////////////////////////////////////////////////////////////////////////////
//
//	TBaseClient static data
//

///	a utility stream....
template< typename OnRecieveData >
typename TBaseClient<OnRecieveData>::_In* TBaseClient<OnRecieveData>::ms_pIn = NULL; 

template< typename OnRecieveData >
typename TBaseClient<OnRecieveData>::_Out* TBaseClient<OnRecieveData>::ms_pOut = NULL; 

	
///	File or Pipe handle.
template< typename OnRecieveData >
HANDLE 	  TBaseClient<OnRecieveData>::ms_hPipe;

///	Overlapped structure
template< typename OnRecieveData >
OVERLAPPED TBaseClient<OnRecieveData>::ms_OverLapWrt;

///	Event handle for overlapped writes.
template< typename OnRecieveData >
HANDLE     TBaseClient<OnRecieveData>::ms_hEventWrt;
	
/// This holds the name of the server
template< typename OnRecieveData >
char TBaseClient<OnRecieveData>::ms_pServerName[ STRING_SIZE ];

///	This holds the name of the client pipe
template< typename OnRecieveData >
char TBaseClient<OnRecieveData>::ms_pClientName[ STRING_SIZE ];

///	an internal buffer (lets use heap allocated memory)
template< typename OnRecieveData >
char TBaseClient<OnRecieveData>::ms_pOutBuffer [ BUFFER_SIZE ];

///	an internal buffer (lets use heap allocated memory)
template< typename OnRecieveData >
char TBaseClient<OnRecieveData>::ms_pInBuffer  [ BUFFER_SIZE ];
	
///	The function object for recieving data
template< typename OnRecieveData >
OnRecieveData TBaseClient<OnRecieveData>::OnRead; 




/////////////////////////////////////////////////////////////////////////////////////////////
//
//	TBaseClient static functions
//

///	The function object for recieving data
template
<
	typename OnRecieveData
>
VOID __stdcall TBaseClient<OnRecieveData>::ReadPipe(HANDLE *hPipe)
{
	//	local data
	//
	DWORD 		dwBytesRead;              	//	Used for ReadFile()
	DWORD 		dwRetCode;                	//	Used to trap return codes.
	DWORD 		dwLastError;              	//	Used to trap returns from GetLastError.
	HANDLE 		hEventRd;              		//	Event handle for overlapped reads.
	OVERLAPPED  OverLapRd;         			//	Overlapped structure.
	DWORD 		dwBytesTransferred;         //	Bytes transferred in read.

	//	Create and init overlap structure.
	//
	hEventRd = CreateEvent( NULL, TRUE, FALSE, NULL );
	std::memset( &OverLapRd, 0, sizeof(OVERLAPPED) );
	OverLapRd.hEvent = hEventRd;

	do
	{
		//	Read the pipe handle.
		//
		dwRetCode = ReadFile( *hPipe       ,  
							  ms_pInBuffer , 
							  BUFFER_SIZE  , 
							  &dwBytesRead , 
							  &OverLapRd   );
		
		//	Do some error checking.
		//
		if( !dwRetCode )
		{                  
			dwLastError = GetLastError();
			
			//	Check for 3 kinds of errors: IO_PENDING, BROKEN_PIPE, or other.
			//	If Error = IO_PENDING, wait for event handle to signal success.
			//
			if( dwLastError == static_cast<DWORD>(ERROR_IO_PENDING) )
			{
				WaitForSingleObject( hEventRd, (DWORD) -1 );
			}
			else
			{                         
				//	if pipe is broken, tell user and break.
				//
				if( dwLastError == static_cast<DWORD>(ERROR_BROKEN_PIPE) )
				{
					std::cerr << "Error: Connection Broken" << std::endl << "> ";
				}
				else
				{            
					std::cerr << "Error: Read Failed" << std::endl << "> ";
				}
				std::cout.flush();
				break;
			}
		}
		
		//	get the overlapped shit...
		//
		GetOverlappedResult( *hPipe, &OverLapRd, &dwBytesTransferred, FALSE );
		
		//	NULL terminate string.
		//
		ms_pInBuffer[ dwBytesTransferred ] = '\0';
		
		
		//	set the strstreams back to the start of the buffer
		//
		ResetInputBuffer();
		
		//	call the templated read object
		//
		OnRead( ms_pIn->m_Buffer );
				
		std::memset(ms_pInBuffer,0,BUFFER_SIZE);
		

	}  while(1);

	//	When pipe is broken, fuck off...
	//
	ExitThread(0);
}




///	\brief	this function will send the data currently stored in the buffer
///	\return	true if OK
///
template
<	
	typename OnRecieveData
>
bool TBaseClient<OnRecieveData>::Send()
{
	DWORD dwBytesWritten,dwLastError,dwRetCode;
	
	//	write the data over
	//
	dwRetCode = WriteFile( ms_hPipe,
									  ms_pOutBuffer,
									  BUFFER_SIZE,
									  &dwBytesWritten,
									  &ms_OverLapWrt);       
			
	//	check the return code
	//							  
	if( !dwRetCode )
	{
		dwLastError = GetLastError();

		//	if error = IO_PENDING, wait until the event signals success.
		//
		if( dwLastError == ERROR_IO_PENDING )
		{

			WaitForSingleObject( ms_hEventWrt, (DWORD) -1 );
		}
	}
		
	
	//	reset stream
	//
	delete ms_pOut;
	ms_pOut = new _Out(ms_pOutBuffer,BUFFER_SIZE);
	std::memset(ms_pOutBuffer,0,BUFFER_SIZE);
		
	return true;
}



///	\brief	this function disconnects (this) client from the server
///	\return	true if OK
///
template
<
	typename OnRecieveData
>
bool TBaseClient<OnRecieveData>::Disconnect()
{
	CloseHandle( ms_hPipe );
	CloseHandle( ms_hEventWrt );
	delete ms_pOut;
	ms_pOut = NULL;
	delete ms_pIn;
	ms_pIn = NULL;
	return true;
}




///	\brief	this function disconnects (this) client from the server
///	\return	true if OK
///
template
<
	typename OnRecieveData
>
bool TBaseClient<OnRecieveData>::Connect( std::string ClientName, std::string ServerName )
{
    DWORD dwBytesWritten;                 // Used for WriteFile().
    DWORD dwThreadID;                     // Used for CreateThread().
    DWORD dwRetCode=0;
    
	//	make the full pipe name
	//
	std::ostrstream oPipeName(ms_pServerName,STRING_SIZE);
	std::ostrstream oClientName(ms_pClientName,STRING_SIZE);
	
	
	oClientName << ClientName;
	oPipeName << "\\\\" <<  ServerName << "\\PIPE\\test";
	
	//	Do CreateFile() to connect to the named pipe.
	//
	ms_hPipe = CreateFile(  ms_pServerName,
							GENERIC_WRITE   | GENERIC_READ,
							FILE_SHARE_READ | FILE_SHARE_WRITE,
                     		NULL, 
                  	    	OPEN_EXISTING, 
                        	FILE_FLAG_OVERLAPPED, 
                        	NULL);
                                    
	//	Do error checking.
	//
	if ( reinterpret_cast<DWORD>(ms_hPipe) == 0xFFFFFFFF)
	{
		dwRetCode = GetLastError();
                    
		//	This error means pipe wasn't found.
		//
		if( (dwRetCode == ERROR_SEEK_ON_DEVICE) || 
			(dwRetCode == ERROR_FILE_NOT_FOUND) )
		{
			std::cerr << "Error: CANT FIND PIPE \"" 
					  << ms_pServerName 
					  << "\"" 
					  << std::endl << "> ";
			std::cout.flush();
			return false;
		}
		else
		{                           
			//	Flagging unknown errors.
			//
			std::cerr << "Error: UNKNOWN" << std::endl << "> ";
			std::cout.flush();
			return false;
		}
	}
	
	//	Create and init overlapped structure for writes.
	//
	ms_hEventWrt = CreateEvent( NULL, TRUE, FALSE, NULL );
	ms_OverLapWrt.hEvent = ms_hEventWrt;

	//	Write the client name to server.
	//
	dwRetCode = WriteFile( ms_hPipe, 
						   ms_pClientName, 
						   BUFFER_SIZE, 
						   &dwBytesWritten, 
						   &ms_OverLapWrt );

	if( !dwRetCode )                   // Wait on overlapped if need be.
	{
		DWORD lastError = GetLastError();
		
		if( lastError == ERROR_IO_PENDING )
		{
			WaitForSingleObject( ms_hEventWrt, (DWORD) -1 );
		}
	}

	//	Create a thread to read the pipe.
	//
	CreateThread( NULL,
				  0,
				  reinterpret_cast<LPTHREAD_START_ROUTINE>(ReadPipe),
				  static_cast<LPVOID>(&ms_hPipe), 
				  0,
				  &dwThreadID);
	return true;
                     	
}