///////////////////////////////////////////////////////////////////////////////////
///
///
///	\file	ClientPipe.cpp
///	\author	Rob Bateman
///	\brief	this is the implimentation of the CClientPipe class. This class maintains
///			a pipe connection to a connecting client.
///	

///////////////////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include "ClientPipe.h"
#include <windows.h>
#include <cstdlib>
#include <cassert>






////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	ctor
///
CClientPipe::CClientPipe()
	: m_hPipe(NULL),
	m_hEventWt(NULL),
	m_hEventRd(NULL),
	m_bLive(false),
	m_Name("")
{
	std::memset( &m_OverLapWt, 0, sizeof(OVERLAPPED) );
	std::memset( &m_OverLapRd, 0, sizeof(OVERLAPPED) );
}

/// \brief	dtor
CClientPipe::~CClientPipe()
{
}


std::string CClientPipe::PipeName = "\\\\.\\PIPE\\test";



///////////////////////////////////////////////////////////////////////////////////

///
///
/// \brief	a function to read a message from the client
///	\param	data	-	a string into which the message will be placed
/// \return	true if OK
///
bool CClientPipe::Read(std::vector<unsigned char> &data)
{
	//	Input buffer for pipe.
	//
	unsigned long iBytesRead,iRetCode;
	unsigned long iBytesTransRd=0;			//	Bytes transferred by overlapped.

	//	is the client live?
	//
	if(!m_bLive)
	{
		return false;
	}

	//	Read the pipe
	//
	iRetCode = ReadFile (m_hPipe, m_ReadBuffer, PLEASE_READ, &iBytesRead, &m_OverLapRd);

	//	is the client still alive?
	//
	if(!m_bLive)
	{
		return false;
	}

	//	Check for three kinds of errors:
	//	If Error = IO_PENDING, wait til
	//	the event handle signals success,
	//	If BROKEN_PIPE, exit the do loop.
	//	Any other error, flag it to the
	//	user and exit the do loop.
	//
	if( !iRetCode )
	{
		switch( GetLastError() )
		{
		//	IO_PENDING, wait on the event.
		//
		case ERROR_IO_PENDING:
			WaitForSingleObject( m_hEventRd, (DWORD)-1 );
			break;

		//	Pipe is broken, exit the loop.
		//
		case ERROR_BROKEN_PIPE:
			return false;
		
		//	Something else is wrong, exit the
		//	the loop after telling the user.
		//
		default:
			std::cerr << "The Pipe has gone mad captain" << std::endl << "> ";
			return false;
		}
	}

	//	get the overlapped result
	//
	GetOverlappedResult( m_hPipe, &m_OverLapRd, &iBytesTransRd, false );

	//	check for a data transfer
	//
	if( iBytesTransRd )
	{
		data.resize( iBytesRead );
		std::memcpy( &data[0], m_ReadBuffer, iBytesRead );
	}

	return true;
}






///////////////////////////////////////////////////////////////////////////////////

///
///
/// \brief	a function to read a message from the client
///	\param	data	-	a string into which the message will be placed
/// \return	true if OK
///
bool CClientPipe::Read(std::string &data)
{
	unsigned long iBytesRead,iRetCode;
	unsigned long iBytesTransRd=0;//	Bytes transferred by overlapped.

	//	is the client live?
	//
	if( !m_bLive )
	{
		return false;
	}

	//	Read the pipe
	//
	iRetCode = ReadFile( m_hPipe      , 
						 m_ReadBuffer , 
						 PLEASE_READ  ,
						 &iBytesRead  , 
						 &m_OverLapRd  );

	//	is the client still alive?
	//
	if( !m_bLive )
	{
		return false;
	}


	data = "";

	//	Check for three kinds of errors:
	//	If Error = IO_PENDING, wait til
	//	the event handle signals success,
	//	If BROKEN_PIPE, exit the do loop.
	//	Any other error, flag it to the
	//	user and exit the do loop.
	//
	if (!iRetCode)
	{
		switch (GetLastError())
		{
		//	IO_PENDING, wait on the event.
		//
		case ERROR_IO_PENDING:
			{
				WaitForSingleObject( m_hEventRd, (DWORD)-1 );
			}	
			break;

		//	Pipe is broken, exit the loop.
		//
		case ERROR_BROKEN_PIPE:
			{
				return false;
			}
		
		//	Something else is wrong, exit the
		//	the loop after telling the user.
		//
		default:
			{
				std::cerr << "The Pipe has gone mad captain" << std::endl << "> ";
				return false;
			}
		}
	}

	//	get the overlapped result
	//
	GetOverlappedResult( m_hPipe, 
						 &m_OverLapRd, 
						 &iBytesTransRd, 
						 false );

	//	check for a data transfer
	//
	if( iBytesTransRd )
	{
		data = m_ReadBuffer;
	}

	return true;
}






///////////////////////////////////////////////////////////////////////////////////

///
///
/// \brief	a function to write a message to the client
///	\param	data	-	the message to send
/// \return	true if OK
///
bool CClientPipe::Write(std::vector<unsigned char> &data)
{
	//	If client isn't alive, don't waste
	//	time writing to it. Shouldn't be, but don't really know
	//
	if (m_bLive)
	{
		//	Used in WriteFile().
		//
		unsigned long iBytesWritten;
		
		
		//	Write the message to the client app
		//
		bool bStatus = static_cast<bool>(WriteFile(  m_hPipe,	
										 			 &data[0],
									 				 static_cast<DWORD>(data.size()),
										  			 &iBytesWritten,	
										  			 &m_OverLapWt ));

		//	check error status
		//
		if( !bStatus )
		{
			unsigned long iLastError = GetLastError();

			//	IO_PENDING, wait on event handle.
			//
			if( iLastError == ERROR_IO_PENDING )
			{
				WaitForSingleObject( m_hEventWt, (DWORD)-1 );
			}
			else
			{
				//	If not NO_DATA, flag user.
				//
				if (iLastError != ERROR_NO_DATA)
				{
					std::cerr << "ERROR: " << GetLastError() << std::endl << "> ";
				}

				//	flag the connections demise
				//
				m_bLive = false;
			}
		}
	}

	return m_bLive;
}
	
	
	
	

///////////////////////////////////////////////////////////////////////////////////

///
///
/// \brief	a function to read a message from the client
///	\param	data	-	a text string to send across the network...
/// \return	true if OK
///
bool CClientPipe::Write(std::string &data)
{
	//	If client isn't alive, don't waste
	//	time writing to it. Shouldn't be, but don't really know
	//
	if (m_bLive)
	{
		//	Used in WriteFile().
		//
		unsigned long iBytesWritten;

		//	Write the message to the client app
		//
		bool bStatus = static_cast<bool>(WriteFile(  m_hPipe,	
										 			 data.c_str(),
									 				 static_cast<DWORD>(data.size()),
										  			 &iBytesWritten,	
										  			 &m_OverLapWt ));

		//	check error status
		//
		if( !bStatus )
		{
			unsigned long iLastError = GetLastError();

			//	IO_PENDING, wait on event handle.
			//
			if( iLastError == ERROR_IO_PENDING )
			{
				WaitForSingleObject( m_hEventWt, (DWORD) -1 );
			}
			else
			{
				//	If not NO_DATA, flag user.
				//
				if( iLastError != ERROR_NO_DATA )
				{
					std::cerr << "ERROR: " 
							  << GetLastError() 
							  << std::endl << "> ";
				}

				//	flag the connections demise
				//
				m_bLive = false;
			}
		}
	}

	return m_bLive;
}






///////////////////////////////////////////////////////////////////////////////////

///
///
///	\brief	this function closes the pipe
/// \return	true if OK
//
bool CClientPipe::ClosePipe()
{
	m_bLive = false;
	CloseHandle( m_hPipe );                // Close handles.
	CloseHandle( m_hEventRd );
	CloseHandle( m_hEventWt );
	DisconnectNamedPipe( m_hPipe );        // Close pipe instance.
	
	return true;
}
	
	
	
	
	

///////////////////////////////////////////////////////////////////////////////////

///
///
///	\brief	this function creates the pipe and sets it waiting for 
///			a client connection. Once called a seperate listening thread
///			is created that waits for a client. When it recieves one, it 
///			attaches itself to the client and creates a new thread that
///			sits listening. Essentually each instance of this class that
///			is created is likely to be associated with it's own thread and
///			it's own client. 
/// \return	true if OK
///
bool CClientPipe::CreatePipe()
{	
	SECURITY_ATTRIBUTES     sa;
	PSECURITY_DESCRIPTOR    pSD;

	//	create a security NULL security
	//	descriptor, one that allows anyone
	//	to write to the pipe... WARNING
	//	entering NULL as the last attribute
	//	of the CreateNamedPipe() will
	//	indicate that you wish all
	//	clients connecting to it to have
	//	all of the same security attributes
	//	as the user that started the
	//	pipe server.
	//
	pSD = static_cast<PSECURITY_DESCRIPTOR>( LocalAlloc(LPTR,SECURITY_DESCRIPTOR_MIN_LENGTH) );

	//	check the psd
	//
	if ( !pSD )
	{
		//	report malloc failure
		//
		std::cerr << "Error in LocalAlloc for pSD" 
				  << std::endl << "> ";
		return false;
	}

	//	initialise the security descriptor structure
	//
	if ( !InitializeSecurityDescriptor( pSD, SECURITY_DESCRIPTOR_REVISION ) )
	{
		//	print error string - unable to initialise descriptor
		//
		std::cerr << "Error: InitializeSecurityDescriptor() "
				  << GetLastError() 
				  << std::endl << "> ";
			
		//	free the security descriptor
		//	  
		LocalFree( static_cast<HLOCAL>(pSD) );
		return false;
	}


	//	add a NULL disc. ACL to the
	//	security descriptor.
	//
	if ( !SetSecurityDescriptorDacl(pSD, TRUE, (PACL) NULL, FALSE) )
	{
	
		std::cerr << "Error: SetSecurityDescriptorDacl() " 
				  << GetLastError() 
				  << std::endl << "> ";
				  
		LocalFree( static_cast<HLOCAL>(pSD) );
		return false;
	}

	//	set the security attributes for the connection
	//
	sa.nLength = sizeof(sa);
	sa.lpSecurityDescriptor = pSD;
	sa.bInheritHandle = TRUE;

	//	Create a local named pipe with
	//	the name '\\.\PIPE\test'.  The
	//	'.' signifies local pipe.
	//
	m_hPipe = CreateNamedPipe (
		PipeName.c_str(),		//	Pipe name = 'test'.
		PIPE_ACCESS_DUPLEX	|		//	2 way pipe.
		FILE_FLAG_OVERLAPPED,		//	Use overlapped structure.
		PIPE_WAIT			|		//	Wait on messages.
		PIPE_READMODE_MESSAGE |  	//	Specify message mode pipe.
		PIPE_TYPE_MESSAGE,
		MAX_PIPE_LIMIT,				//	Maximum instance limit.
		OUT_BUF_SIZE,				//	Buffer sizes.
		IN_BUF_SIZE,
		TIME_OUT,					//	Specify time out.
		&sa);						//	Security attributes.


	//	Check Errors.
	//
	if ( reinterpret_cast<DWORD>(m_hPipe) == 0xFFFFFFFF )
	{
		std::cerr << "Failed to create pipe. This should never happen!" << std::endl;
		return false;
	}


	//	Block until a client connects.
	//
	ConnectNamedPipe( m_hPipe, NULL );


	//	m_hEventRd and init overlap for writing.
	//
	m_hEventWt = CreateEvent( NULL, true, false, NULL );
	m_OverLapWt.hEvent = m_hEventWt;


	//	Create and init overlap for reading.
	//
	m_hEventRd = CreateEvent( NULL, true, false, NULL );
	m_OverLapRd.hEvent = m_hEventRd;


	//	set any other values for the connection
	//
	m_bLive 	= true;


	//	Input buffer for pipe.
	//
	char inBuf[IN_BUF_SIZE] = "";	


	//	Read from the client, the first
	//	first message should always be
	//	the clients user name.
	//
	unsigned long iBytesRead=0,iRetCode;
	
	iRetCode = ReadFile( m_hPipe      , 
						 inBuf        , 
						 PLEASE_READ  , 
						 &iBytesRead  , 
						 &m_OverLapRd );


	//	if failed, get error
	//
	if( !iRetCode )
	{	
		//	if still waiting
		//
		if( GetLastError() == ERROR_IO_PENDING ) 
		{
			WaitForSingleObject( m_hEventRd, (DWORD) - 1 );
		}
	}

	//	store client's name 
	//
	m_Name = inBuf;


	return true;
}




