///////////////////////////////////////////////////////////////////////////////////
///
///
///	\file	ClientPipe.h
///	\author	Rob Bateman
/// \brief	this file defines the class CClientPipe which manages a single pipe between
///			the server and a client. An instance of this class will exist for every client
///			connection which will maintain a thread

///////////////////////////////////////////////////////////////////////////////////

#pragma once

#ifndef _CLIENT_PIPE_H_
#define _CLIENT_PIPE_H_


///////////////////////////////////////////////////////////////////////////////////


#include <windows.h>
#include <iostream>
#include <string>
#include <vector>
#include <list>


///////////////////////////////////////////////////////////////////////////////////


//	symbolic constants. 64Kb buffers & packets
//
const unsigned int IN_BUF_SIZE     =   65536;
const unsigned int OUT_BUF_SIZE    =   65536;
const unsigned int TIME_OUT        =   0   ;
const unsigned int PLEASE_READ     =   65536;
const unsigned int PLEASE_WRITE    =   65536;
const unsigned int MAX_PIPE_LIMIT  =   100;


///////////////////////////////////////////////////////////////////////////////////

/// \brief	this class is designed to hold a reference to a connected client app.
///
class CClientPipe
{

friend class CServer;

private:

	static std::string PipeName;

///////////////////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////////////
	//
	//	Pipe handles and such like.... 
	//

	/// a handle to the pipe
	HANDLE      m_hPipe;
	///	a handle to the read event
	HANDLE      m_hEventRd;
	///	the overlapped read
	OVERLAPPED  m_OverLapRd;
	///	a handle to the write event
	HANDLE      m_hEventWt;
	/// the overlapped write
	OVERLAPPED  m_OverLapWt;



	///////////////////////////////////////////////////////////////////////////////
	//
	//	Info about the client
	//
	
	/// the name of the client
	std::string m_Name;
	/// is the connection live
	bool        m_bLive;
	
	
	
	///////////////////////////////////////////////////////////////////////////////
	//
	//	Internal Buffers for reading/writing. 
	//
	
	///	a buffer for reading
	char m_ReadBuffer[IN_BUF_SIZE];
	///	a buffer for writing
	char m_WriteBuffer[OUT_BUF_SIZE];


public:

	///////////////////////////////////////////////////////////////////////////////

	/// \brief	ctor
	CClientPipe();

	/// \brief	dtor
	~CClientPipe();
	

	///////////////////////////////////////////////////////////////////////////////	

	/// \brief	a function to read a message from the client
	///	\param	data	-	a string into which the message will be placed
	/// \return	true if OK
	///
	bool Read(std::vector<unsigned char> &data);
	
	/// \brief	a function to read a message from the client
	///	\param	data	-	a string into which the message will be placed
	/// \return	true if OK
	///
	bool Read(std::string &data);

	/// \brief	a function to write a message to the client
	///	\param	data	-	the message to send
	/// \return	true if OK
	///
	bool Write(std::vector<unsigned char> &data);
	
	/// \brief	a function to read a message from the client
	///	\param	data	-	a text string to send across the network...
	/// \return	true if OK
	///
	bool Write(std::string &data);
	
	

	///////////////////////////////////////////////////////////////////////////////	

	///	\brief	this function closes the pipe
	/// \return	true if OK
	//
	bool ClosePipe();
	
	

	///////////////////////////////////////////////////////////////////////////////	
	
	///	\brief	this function creates the pipe and sets it waiting for 
	///			a client connection. Once called a seperate listening thread
	///			is created that waits for a client. When it recieves one, it 
	///			attaches itself to the client and creates a new thread that
	///			sits listening. Essentually each instance of this class that
	///			is created is likely to be associated with it's own thread and
	///			it's own client. 
	/// \return	true if OK
	///
	bool CreatePipe();
	
	

	///////////////////////////////////////////////////////////////////////////////	

	///	\brief	this function kills the connection to the client
	///	
	inline void KillConnection() {
		m_bLive = false;
	}
	
	

	///////////////////////////////////////////////////////////////////////////////	

	///	\brief	a function to see if the client connection is alive
	/// \return	true if a client is connected
	///
	inline bool IsAlive() const {
		return m_bLive;
	}
	
	

	///////////////////////////////////////////////////////////////////////////////	

	///	\brief	this function returns the name of the client 
	///	\return	the clients name
	///
	inline const std::string Name() {
		return m_Name;
	}
};



#endif