////////////////////////////////////////////////////////////////////////////////////

/// 

#define WIN32_LEAN_AND_MEAN
#include "server.h"
#include <cstring>
#include <cstdio>
#include <cstdlib>





////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	This function forms the main server thread entry point
///	\return	the thread exit code
///
DWORD __stdcall CServer::ServerProc(void* ptr)
{
	UNREFERENCED_PARAMETER(ptr);
	return static_cast<DWORD>( CServer::Run() );
}





////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	A function to forcably remove a connected client from the server
///	\param	client	-	the client to remove 
///	\return	true if successfully kicked
///
bool CServer::Kick(std::string client)
{
	for( std::list<CClientPipe*>::iterator it = m_Connections.begin() ; it != m_Connections.end(); ++it )
	{
		if((*it)->Name() == client)
		{
			(*it)->KillConnection();
			Tell(client,"");
			delete (*it);
			m_Connections.erase(it);
			return true;
		}
	}
	return false;
}







////////////////////////////////////////////////////////////////////////////////////

///	a list of all connections to this client
std::list<CClientPipe*>  CServer::m_Connections;
	
/// is the server running?
bool CServer::m_bRunning = false;

/// the thread id of the first thread.
unsigned long CServer::lpServerThreadID = 0;




////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	oh dear. This function is fairly important!. In order to kill the server,
///			we need to break all client connections to the server in such a way as to 
///			force the thread to exit without causing massive errors. 
///
///			One thing we could do would be to close all the pipes and then delete the client
///			connections. This would be bad. The seperate threads may be in the process of reading
///			or writing data to the clients. Killing the pipe mid process without letting the thread
///			know would mean the thread would access volatile memory and problably crash. To prevent
///			this happening, we set a flag within the ClientPipe instance that the thread checks to
///			see if the pipe should be running. If it is false, the thread will close the pipe for us
///			and then gracefully exit.
///
///			The problem is that having set the flags we need to make sure that the threads have exited
///			so that we can delete the client pipe from our list. For this we can simply send a message
///			to all clients so that they check the exit flags and gracefully quit. Sleep for a few milliseconds
///			to be sure and then erase all of the pipes.
///
void CServer::KillServer() 
{
	//	set flags in each CClientPipe class to tell the thread to exit
	//
	for( std::list<CClientPipe*>::iterator it = m_Connections.begin() ; it != m_Connections.end(); ++it )
	{
		(*it)->KillConnection();
	}

	//	use tell all to force an update of the threads
	//
	CServer::TellAll("");

	//	wait for a sec to ensure all threads have quit
	//
	Sleep(1000);
	
	
	//	delete all instances of CClientPipe classt
	//
	for( std::list<CClientPipe*>::iterator it = m_Connections.begin() ; it != m_Connections.end(); ++it )
	{
		delete (*it);
	}


	//	delete the connections
	//
	m_Connections.erase(m_Connections.begin(),m_Connections.end());

	//	set the flag in the server to quit.
	//
	m_bRunning = false;
}





////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	returns the number of connected clients
///	\return	the number of connected clients
///
unsigned int CServer::GetNumClients()  
{
	//	the list will always be 1 bigger than the number of clients (because one pipe 
	//	will be listening for a new connection)
	//
	return static_cast<unsigned int>(m_Connections.size())-1;
}





////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	a function to get a list of connected client names
///	\param	the array to be filled with the client names
///
void CServer::GetClientNames( std::vector<std::string> &NameList )
{
	//	ensure that the array is empty
	//
	NameList.erase(NameList.begin(),NameList.end());

	//	iterate through all connections
	//
	std::list<CClientPipe*>::iterator it = m_Connections.begin();
	for(;it!=m_Connections.end();++it)
	{
		//	add the name to the back
		//
		NameList.push_back((*it)->Name());
	}
}






////////////////////////////////////////////////////////////////////////////////////

/// 
///	init the server
///
bool CServer::InitServer(std::string PipeName) 
{
	m_bRunning = true;
	
	CClientPipe::PipeName = "\\\\.";
	CClientPipe::PipeName += PipeName;

	//	Create the first instance of a server side of the pipe.
	//
	CreateThread (
		static_cast<LPSECURITY_ATTRIBUTES>(NULL),  		//	No security attributes.
		static_cast<DWORD>(0),                         	//	Use same stack size.
		static_cast<LPTHREAD_START_ROUTINE>(ServerProc),//	Thread procedure.
		static_cast<LPVOID>(NULL),                     	//	Parameter to pass.
		static_cast<DWORD>(0),                         	//	Run immediately.
		static_cast<LPDWORD>(&lpServerThreadID) );      	//	Thread identifier.
		
	return true;	
};






////////////////////////////////////////////////////////////////////////////////////

/// 
///	\brief	this function is only called by ServerProc. It creates a pipe that
///			listens for a connection. When a connection is made, the 
/// \return the thread return code
///
DWORD CServer::Run()
{
	char					inBuf[IN_BUF_SIZE] = "";	//	Input buffer for pipe.
	unsigned long			iServerThreadID;			//	Used for CreateThread().

	//	create an iterator to a new client in the list
	//
	std::list<CClientPipe*>::iterator	it = m_Connections.insert(m_Connections.end(),new CClientPipe());

	if( (*it) )
	{
		//	create the pipe
		//
		(*it)->CreatePipe();


		//	Create a thread which will make
		//	another server instance of the
		//	named pipe.
		//
		CreateThread (
			static_cast<LPSECURITY_ATTRIBUTES>(NULL),  		//	No security attributes.
			static_cast<DWORD>(0),                         	//	Use same stack size.
			static_cast<LPTHREAD_START_ROUTINE>(ServerProc),//	Thread procedure.
			static_cast<LPVOID>(NULL),                     	//	Parameter to pass.
			static_cast<DWORD>(0),                         	//	Run immediately.
			static_cast<LPDWORD>(&iServerThreadID) );      	//	Thread identifier.
			

		//	forces a redraw of all clients...
		//
		CServer::TellAll("");

		//	Do loop which basically reads from
		//	this specific client, and then
		//	uses TellAll() to broadcast the
		//	message to all the connected
		//	clients.
		//
		std::string s;
		while( (*it)->Read(s) )
		{
			CServer::TellAll(s);
		}


		//	close the pipe down
		//
		(*it)->ClosePipe();
		
		
		//	delete the pipe
		//
		delete (*it);

		//	delete connection from list
		//
		m_Connections.erase(it);
	
	}
	else //  if mem allocation failed....
	{
		std::cerr << "Memory allocation failure" << std::endl;
		std::exit(0);
	}

	return 0;
}







////////////////////////////////////////////////////////////////////////////////////

/// 
//	This procedure sends the buffer argument to each of the clients
//
void CServer::TellAll( std::string buffer )
{
	//	prints the message so we can track what the server is doing
	//
	std::cout << "Sending Message:\n> \"" << buffer << "\"\n> ";

	//	iterate through all current client connections.
	//
	for( std::list<CClientPipe*>::iterator it = m_Connections.begin() ; it != m_Connections.end(); ++it )
	{
		//	send the string to the client
		//
		(*it)->Write(buffer);	
	}
}





////////////////////////////////////////////////////////////////////////////////////

/// 
//	This procedure sends the buffer argument to each of the clients
//
void CServer::Tell( std::string user, std::string msg )
{
	//	prints the message so we can track what the server is doing
	//
	std::cout << "Sending Message:\n> \"" << msg << "\" to \"" << user << "\"\n> ";

	//	iterate through all current client connections.
	//
	for( std::list<CClientPipe*>::iterator it = m_Connections.begin() ; it != m_Connections.end(); ++it )
	{
		if((*it)->Name() == user) {
			//	send the string to the client
			//
			(*it)->Write(msg);
			return; 
		}
	}
}














