// /////////////////////////////////////////////////////////////////////////////////////////
///
///
///	\file	Server.h
///	\author	Rob Bateman
/// \brief	this file defines the class CServer which manages all client connections.
///
// /////////////////////////////////////////////////////////////////////////////////////////



// /////////////////////////////////////////////////////////////////////////////////////////
//
//	Sentry

#pragma once

#ifndef SERVER_H_
#define SERVER_H_



// /////////////////////////////////////////////////////////////////////////////////////////
//
//	Include

#include "ClientPipe.h"

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#include <windows.h>
#include <iostream>
#include <string>
#include <vector>
#include <list>



// /////////////////////////////////////////////////////////////////////////////////////////
///
/// \brief	a class to manage all client connections. 
///	
// 
class CServer
{
private:

	///	\brief	the server proc is required to be a friend of the server
	///
	static DWORD __stdcall ServerProc(void*);

public:

	///	\brief	this function starts the server
	///	\param	PipeName - the name of the piped connection
	///
	static bool InitServer(std::string PipeName);


	///	\brief	oh dear. This function is fairly important!. In order to kill the server,
	///			we need to break all client connections to the server in such a way as to 
	///			force the thread to exit without causing massive errors. 
	///
	///			One thing we could do would be to close all the pipes and then delete the client
	///			connections. This would be bad. The seperate threads may be in the process of reading
	///			or writing data to the clients. Killing the pipe mid process without letting the thread
	///			know would mean the thread would access volatile memory and problably crash. To prevent
	///			this happening, we set a flag within the ClientPipe instance that the thread checks to
	///			see if the pipe should be running. If it is false, the thread will close the pipe for us
	///			and then gracefully exit.
	///
	///			The problem is that having set the flags we need to make sure that the threads have exited
	///			so that we can delete the client pipe from our list. For this we can simply send a message
	///			to all clients so that they check the exit flags and gracefully quit. Sleep for a few milliseconds
	///			to be sure and then erase all of the pipes.
	///
	static void KillServer() ;


	///	\brief	a function to send a message to all connected clients
	///	\param	s	-	the message text
	///	
	static void TellAll(std::string);

	///	\brief	send a message to the specified user
	///	\param	client	-	the name of the client that is to recieve the message
	///	\param	msg	-	the message text
	///
	static void Tell(std::string client, std::string msg);

	///	\brief	a function to forcably disconnect (kick) a client
	///	\param	client	-	the name of the client to kick
	///	\return	true if OK
	///
	static bool Kick(std::string client);

	///	\brief	returns the number of connected clients
	///	\return	the number of connected clients
	///
	static unsigned int GetNumClients();

	///	\brief	a function to get a list of connected client names
	///	\param	the array to be filled with the client names
	///
	static void GetClientNames( std::vector<std::string> &NameList );

	///	\brief	a function to see if the server is running
	///	\return	true if server is active
	///
	static bool IsRunning(){
		return m_bRunning;
	}


protected:

	///	a list of all connections to this client
	static std::list<CClientPipe*>  m_Connections;


private:
	
	///	\brief	this function is only called by ServerProc. It creates a pipe that
	///			listens for a connection. When a connection is made, the 
	/// \return the thread return code
	///
	static DWORD Run();
	
	/// is the server running?
	static bool m_bRunning;

	/// the thread id of the first thread.
	static unsigned long lpServerThreadID;
};



///	\brief	this is the default Server sentinal check. Provide your
///			own versions for RunServerConsole to use.
///
struct ServerRunningCheck 
{
	bool operator()() 
	{
		return CServer::IsRunning();
	};
}; 



///	\brief	this template is designed to aid making a command line based server
///
template
<
	typename Action ,	//	function object to perform the on-recieve action
	typename Sentinal	//	function object for the terminating condition 
	
	///	code warrior doesn't support default template args
	///
	#if( _MSC_VER_MSC_VER >= 800 || defined(__BORLANDC__) )
		= ServerRunningCheck
	#endif
>
void RunServerConsole () 
{
	//	start the server
	if( !CServer::InitServer( "\\PIPE\\test" ) ) {
		std::cout << "Failed to init the server" << std::endl;
		return;
	}
		
	//	buffer for user input
	char buff[512] = {0};
	
	//	lets go DOS!
	std::cout << "> ";
	
	//	keep track of what was last read in
	int i = 0;
	
	//	function objects
	Sentinal _sentinal;
	Action   _action;
	
	//	Loop until sentinal returns false
	while( _sentinal() )
	{
		//	grab a byte of text
		std::cin.get( buff[i] );
		
		//	if newline,
		if( buff[i] == '\n' )
		{
			//	terminate the string		
			buff[i] = 0;
			
			//	perform the template action
			_action(buff);
			
			//	zero memory and reset count
			std::memset(buff,0,512);
			i=0;
			
			//	lets go DOS!
			std::cout << "> ";
		}
		else
		{
			//	move to next char
			++i;
		}
	}
}


//	this is for code-warrior. It does not support default template
//	function args so I use partial-specialisation instead
//
#if ( !(_MSC_VER_MSC_VER >= 800 || defined(__BORLANDC__)) )

///	\brief	this template is designed to aid making a command line based server
///
template
<
	typename Action 
>
void RunServerConsole () 
{
	RunServerConsole< Action , ServerRunningCheck >();
}	

#endif


#endif